import System.IO as Sys

import System.Exit(die)
import System.Environment(getArgs, getProgName)
import Control.Parallel.Strategies (parBuffer, using, rseq, rpar, parList, rdeepseq, rparWith)
import Data.Char
import System.Posix.IO
import System.Posix.Types
import Data.Set

-- A board is represented as a list of lists of characters
type Board = [[Char]]

makeBoard :: String -> Int -> Board
makeBoard [] _ = []
makeBoard input dim = Prelude.take dim input : makeBoard (Prelude.drop dim input) dim

-- | Sequential implementation
seqWordHuntSolver :: Board -> Set String -> [String]
seqWordHuntSolver board dict = 
  [word | (x,y) <- indices, 
          word <- findWords (x,y) [] "", 
          word `member` dict,
          length word > 2]
  where
    indices = [(x,y) | x <- [0..(length board - 1)], 
                       y <- [0..(length (head board) - 1)]]
 
    findWords :: (Int, Int) -> [(Int, Int)] -> String -> [String]
    findWords (x,y) visited word
      | x < 0 || y < 0 || x >= length board || y >= length (head board) || (x,y) `elem` visited = [] -- base case
      | otherwise = 
          let newWord = word ++ [board !! x !! y]
              newVisited = (x,y) : visited
          in  [newWord] ++ 
              (findWords (x-1,y-1) newVisited newWord ++
               findWords (x-1,y)   newVisited newWord ++
               findWords (x-1,y+1) newVisited newWord ++
               findWords (x,y-1)   newVisited newWord ++
               findWords (x,y+1)   newVisited newWord ++
               findWords (x+1,y-1) newVisited newWord ++
               findWords (x+1,y)   newVisited newWord ++
               findWords (x+1,y+1) newVisited newWord)

-- | Parallel implementation
wordHuntSolver :: Board -> Set String -> [String]
wordHuntSolver board dict = 
  [word | word <- concat allWords, 
          word `member` dict,
          length word > 2]
  where
    -- Find all the indices (row and column) of the squares on the board
    indices = [(x,y) | x <- [0..(length board - 1)], 
                       y <- [0..(length (head board) - 1)]]

    parFindWords = findWords [] ""
    allWords = Prelude.map parFindWords indices `using` parBuffer 2 rdeepseq  -- reduced to weak head normal form, rpar, deepseq instead
    -- allWords' = rdeepseq allWords

    -- Find all the words that can be formed starting at a given square
    -- and following a path of adjacent squares
    findWords :: [(Int, Int)] -> String -> (Int, Int) -> [String]
    findWords visited word (x,y)
      -- If the current square is out of bounds or has already been visited,
      -- there are no more words to be found
      | x < 0 || y < 0 || x >= length board || y >= length (head board) || (x,y) `elem` visited = []
      -- Otherwise, add the current square to the visited squares, add its
      -- character to the current word, and search for more words in all
      -- the adjacent squares
      | otherwise = 
          let newWord = word ++ [board !! x !! y]
              newVisited = (x,y) : visited
          in  [newWord] ++ 
              (findWords newVisited newWord (x-1,y-1) ++
               findWords newVisited newWord (x-1,y)   ++
               findWords newVisited newWord (x-1,y+1) ++
               findWords newVisited newWord (x,y-1)   ++
               findWords newVisited newWord (x,y+1)   ++
               findWords newVisited newWord (x+1,y-1) ++
               findWords newVisited newWord (x+1,y)   ++
               findWords newVisited newWord (x+1,y+1)) 


readDictionary :: FilePath -> IO [String]
readDictionary path = lines <$> readFile path

parseDict :: [[Char]] -> Set [Char]
parseDict dictionary = Data.Set.fromList (Prelude.map (Prelude.map toLower) dictionary)

main :: IO ()
main = do 
         args <- getArgs
         case args of 
            [dict, board, dim] -> do
               dictionary <- readDictionary dict
               let parsed =  parseDict dictionary
                   solved = wordHuntSolver (makeBoard board (read dim :: Int)) parsed
               mapM_ putStrLn solved 
            _ -> 
               do pn <- getProgName
                  die $ "Usage: " ++pn++ " <dictionary-filename> <board> <dimension>"


-- | Test for DFS (wordHuntSolver)
test :: IO ()
test = do
  let board = [
          "abcd",
          "efgh",
          "ijkl"
        ]

      dict = fromList ["a", "bef", "abe", "fgk", "jie", "goodness", "kgfb", "efg", "hello", "fkplhg"]
      -- expected = ["abe", "bef", "efg", "fgk", "jie", "kgfb", "hello"]
      expected = ["a", "bef", "abe", "fgk", "jie", "goodness", "kgfb", "efg", "hello", "fkplhg"]

  -- Check that the wordHuntSolver function returns the expected result
  assertEqual (wordHuntSolver board dict) expected

-- Assert that two values are equal
assertEqual :: (Eq a, Show a) => a -> a -> IO ()
assertEqual x y
  | x == y    = return ()
  | otherwise = error (show x ++ " /= " ++ show y)



-- | My attempt at the second part of the parallelization
readChunks :: Fd -> IO [String]
readChunks fd = do
  -- fileSize <- hFileSize dict
  -- let fileMode   = Just (CMode 0440)
  --     part = fileSize `div` 8
  --     part' = part * parts
  --     size = if parts == (totalParts-1)
  --               then part
  --               else part + 
  chunk <- fdRead fd 128--4096
  done <- isEOF
  -- if fst chunk == ""
  if done
    then return []
    else do
      rest <- readChunks fd
      -- putStrLn (fst chunk)
      return ( fst chunk : rest)


